#include <ESP8266WiFi.h>
#include <WiFiClient.h>
#include <EEPROM.h>
#include <ESP8266WebServer.h>
#include <WiFiUdp.h>
#include <functional>
#include "switch.h"
#include "UpnpBroadcastResponder.h"
#include "CallbackFunction.h"
#include "Access_Point.h"
 // ***********************************************************************************
IPAddress timeServer(129, 6, 15, 28); // time.nist.gov NTP server
// IPAddress timeServerIP; // time.nist.gov NTP server address
const char* ntpServerName = "time.nist.gov";
unsigned int localPort = 2390;      // local port to listen for UDP packets
const int NTP_PACKET_SIZE = 48; // NTP time stamp is in the first 48 bytes of the message
const int Time_Zone = 2;        // Time zone adjustment for NTP time server
byte packetBuffer2[ NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets 
const byte NTP_Wait_Time = 100; // time in multiples of 10 mS that device will wait 
WiFiUDP udp;          // A UDP instance for NTP clock

  //----------------------- Alexa --------------------------
void SwitchOneOn();     // Relay 1
void SwitchOneOff();
void SwitchTwoOn();     // Relay 2
void SwitchTwoOff();
void SwitchThreeOn();   // Relay 3
void SwitchThreeOff();
boolean wifiConnected = false;
UpnpBroadcastResponder upnpBroadcastResponder;
Switch *SwitchOne = NULL;
Switch *SwitchTwo = NULL;
Switch *SwitchThree = NULL;
  // ------------------------------------------------------------------
boolean NTP_Timeout;   // if true, NTP attempt timed out

char KeyboardBuffer[Max_String_Length];      // Keyboard buffer
 // -------------------- LED timer and flash rates ----------------------------------
const unsigned long LED_On_Flash = 1000;   // Station mode
const unsigned long LED_Off_Flash = 150;   
const unsigned long NTP_Not_Set_On = 150; // NTP clock update  is not resolved
const unsigned long NTP_Not_Set_Off = 1000;
const unsigned long LED_Flash_On_Lost_Connection = 50;
const unsigned long LED_Flash_Off_Lost_Connection = 300;
const unsigned long LED_Flash_On_Startup = 50;
const unsigned long LED_Flash_Off_Startup = 50;
unsigned long LED_On_Timer = LED_Flash_On_Startup;   // flashing LED ON timer with Start up (mS) flash rates
unsigned long LED_Off_Timer = LED_Flash_Off_Startup;  // flashing LED OFF timer with start up (mS) flash rate
 // ------------------ PUMP PROTECTION -------------------------
const boolean Buzzer_On = LOW;   // IO state for buzzer on   
const boolean Buzzer_Off = !Buzzer_On; 
boolean Alarm = false;
 // ------------------------------------------------------------
unsigned long Loop_Timer;     // used as timer for timed operations. Eg LED flashing
unsigned long NTP_Timer;      // used as timer for the NTP time interval

const unsigned long One_Minute_timer_Tick_Threshold = 60000; // used to determine one minute timer activity (mS)

const byte Login_Timeout = 100;  // maximum number of log in attempts before forced reset
byte Login_Attempts_Counter;  // Counter for for of attempted logins at any one go.

byte Command = 255;       // input buffer for main screen commands

boolean PIR_Switch_Flag[] = {0,0,0}; // flag = true = PIR1,PIR2 or PIR3 initiated a lamp switch

boolean Timer_Active[] = {false,false,false};   // for each relay, true = a time control timer (timer 1 and/or timer 2)is running ON
byte Lost_Connection_Timer;   // minutes timer if no connection - will force reset on time out
boolean Access_Point_Running;   // true = AP is running

const boolean AP_Switch_On = LOW;   // IO status for AP switch ON
const boolean AP_Switch_Off = !AP_Switch_On;
const byte Lost_Connection_Timeout = 20;    // device will reset after 20 minutes of lost connection

// --------------------------------------------------------
void setup()
{
  Serial.begin(SERIAL_BAUDRATE);
  EEPROM.begin(512);            // Enable 512 bytes of EEPROM
  Alarm = false;
  Auto_EEPROM_Check();   // check to see if EEPROM is initialised, if not set to defaults
  No_Of_Relays = EEPROM.read(EEPROM_No_Of_Relays);  // set number of relays on the board
  No_Of_AUX = EEPROM.read(EEPROM_No_Of_AUX);  
  if (No_Of_Relays >3 || No_Of_Relays == 0) No_Of_Relays = 1;  // precaution - ensure valid number is present
  if (No_Of_AUX >3) No_Of_AUX = 0;  
  byte i = 0; 
  while (i < No_Of_AUX)
  {
    pinMode(AUX_Pin[i],INPUT_PULLUP); // initialise AUX input pins
    i++;
  } 
  pinMode(AP_Switch, INPUT_PULLUP);   
  i = 0;
  while (i < No_Of_Relays)
  {  
    pinMode(Relay_Pin[i], OUTPUT);           // Initialise Relay output pins
    digitalWrite(Relay_Pin[i], Relay_Off);   // set relay OFF
    i++;
  }
  pinMode(Heartbeat_LED,OUTPUT);  // Inititialise Heart beat LED   
  Serial.println("");
  Serial.println(Firmware_Version);   // display firmware revision number
  Serial.println(F("Developer: Chris Gimson     chris@gimson.co.uk"));
  Enable_Messages = EEPROM.read(EEPROM_Enable_Messages);  // setthe display messages flag 
  Soft_AP_Switch = false;    // Access point is disabled
  if (Enable_Messages)       // If display messages is enabled
  {
     Serial.println("");
     Serial.println(Firmware_Version);
     Serial.println(F("Display messaging is enabled"));
     Serial.print(No_Of_Relays);
     Serial.println(F(" Relay(s) fitted"));
  }
  Download_Watchdog_Type_And_Alexa_Enable(false);   // Download type of watchdog in use & which output is Alexi enabled  
  Download_WiFi_Login_Details_From_EEPROM(false);   // Download EEPROM values to working memory
  Download_From_EEPROM_Output_And_Input_Settings(false);
  Download_From_EEPROM_Relay_Descriptors(false); 
  Download_Timer_Reset_Clock_From_EEPROM(false); 
  Download_From_EEPROM_Timer_Control_Parameters(false);  
  pinMode(Buzzer, OUTPUT);   // initialise Relay 2 pin as an output for buzzer if needed (pump protection)
  digitalWrite(Buzzer,Buzzer_Off);
  Login_Attempts_Counter = Login_Timeout; 
  WifiConnect();      // connect to wifi network                      
  NTP_Resolved = false;   // true = soft clock successfully by NTP
  if (WiFi.status() == WL_CONNECTED)  // if the Wifi has successfully connected
  {                                   // set up Alexa functions     
      byte Alexi_Count = 0;
      if (Alexa_Enable[0] == true) Alexi_Count ++;
      if (No_Of_Relays >= 2 && Alexa_Enable[1]== true) Alexi_Count ++; // increment Alexi_Count if relay 2 is fitted AND Alexi enabled
      if( No_Of_Relays == 3 && Alexa_Enable[2]== true) Alexi_Count ++; // increment Alexi_Count if relay 3 is fitted AND Alexi enabled
      if (Alexi_Count > 0) // if any output is Alexa enabled
      {
        if (Enable_Messages) Serial.println("Starting Alexa switches upnp broadcast responder");
        upnpBroadcastResponder.beginUdpMulticast(); 
      }
       // Define switches Format: Alexa invocation name, local port no, on callback, off callback      
      if (Alexa_Enable[0])
      {      // only if Alexa voice control 1 is enabled         
        SwitchOne = new Switch(Relay_Descriptors[0].c_str(), 81, SwitchOneOn, SwitchOneOff);
        upnpBroadcastResponder.addDevice(*SwitchOne);  // for relay 1        
      }
      if(No_Of_Relays >= 2 && Alexa_Enable[1] == true) 
      {
        SwitchTwo = new Switch(Relay_Descriptors[1].c_str(), 82, SwitchTwoOn, SwitchTwoOff);
        upnpBroadcastResponder.addDevice(*SwitchTwo); // if 2 or 3 relays fitted        
      }
      if(No_Of_Relays == 3 && Alexa_Enable[2] == true) 
      {   // 3rd relay is fitted and Alexa enabled - what port to use?
        SwitchThree = new Switch(Relay_Descriptors[2].c_str(), 83, SwitchThreeOn, SwitchThreeOff);       
        upnpBroadcastResponder.addDevice(*SwitchThree); // if 3 relays fitted             
      }   // end of if(No_Of_Relays == 3 && Alexa_Enable[2] == true)
                    // no of enabled Alexa relays is < 3, set up Access point wake up command         
  }   // end of if (WiFi.status() == WL_CONNECTED)  
   NTP_Time_Interval = 10000;   // set up NTP interval timer (mS)
   udp.begin(localPort);   // for the NTP server
   delay(100);
   Loop_Timer = millis();  // for the LED flash rates
   Timer_Tick = millis();  // for the output and PIR timers 
   NTP_Timer = millis();   // for the NTP time interval timer
   Lost_Connection_Timer = 0;
   Access_Point_Running = false;
}    // end of Setup
// -------------------------------------------------------- 
void loop()
{
   byte i = Check_AP_Switch(); 
   if(i == 2) Soft_AP_Switch = !Soft_AP_Switch;  // Toggle if AP switch pressed for > 5 seconds                 
   if(Soft_AP_Switch == true)    // Access Point (AP)
   {   // AP is enabled
      if (Access_Point_Running == false)  // is AP already running?
      {                                   // not running so set up AP
         Set_Up_Access_Point();
         Access_Point_Running = true;    
      }   // end of  if (Access_Point_Running == false) 
      else 
      {
        serverAP.handleClient();  // AP is running so maintain AP web page 
      }
   }  // end of    if(Soft_AP_Switch == true)    // Access Point (AP) 
   else
   {                //  AP is disabled
      WiFi.softAPdisconnect(true);   // disconnect the AP
      serverAP.stop();               // Stop  the AP webserver 
      WiFi.mode(WIFI_STA);           // Switch to station mode only
      Access_Point_Running = false;
      Check_AUX_Button();   // Check to see for manual operation of relays via manual button
      if (No_Of_AUX  > 0)  // only if there is at least one AUX input enabled
      {
         i = 0;
         while (i < No_Of_AUX)  // check PIR inputs for any activity
         {
           Check_PIR_Inputs(i);
           i++;
         }
      }  // end of   if ( > 0)  // only if there is at least one PIR enabled          
   }  // end of  else                 
                    // Station mode  
   if (Login_Successful)
   {  
      if (WiFi.status() == WL_CONNECTED) 
      {
         Lost_Connection_Timer = 0;   // reset the lost connection timer
         if(NTP_Resolved)
         {
            LED_On_Timer = LED_On_Flash;   // Set LED flash rate - Station = connected, with NTP resolved
            LED_Off_Timer = LED_Off_Flash;  
         }
         else
         {
            LED_On_Timer = NTP_Not_Set_On;   // Set LED flash rate for NTP not set up
            LED_Off_Timer = NTP_Not_Set_Off; 
         }
         if (Alexa_Enable[0] || Alexa_Enable[1] || Alexa_Enable[2]) 
         {
            upnpBroadcastResponder.serverLoop();
            if (Alexa_Enable[0]) SwitchOne->serverLoop();  // for relay 1
            if(No_Of_Relays >= 2 && Alexa_Enable[1] == true)SwitchTwo->serverLoop(); // if 2 or 3 relays fitted
            if(No_Of_Relays == 3 && Alexa_Enable[2] == true)SwitchThree->serverLoop();  // if 3 relays enabled 
         }  // end of if (Alexa_Enable[0] || Alexa_Enable[1] || Alexa_Enable[2]) 	        
	    } // end of if (WiFi.status() == WL_CONNECTED)  
      else   // Lost connection
      {
          LED_On_Timer = LED_Flash_On_Lost_Connection;   // Reset LED flash rate - lost connection
          LED_Off_Timer = LED_Flash_Off_Lost_Connection;
      }  
   }  // end of if(Login_Successful)  
   Check_For_Keyboard_Command();  
   unsigned long LoopTimer = millis() - Loop_Timer;   // Grab time value since last scan
   if (digitalRead(Heartbeat_LED) == LED_On)   // If the Heartbeat LED is ON
   {
     if (LoopTimer >= LED_On_Timer) 
     {
       digitalWrite(Heartbeat_LED,LED_Off); // On timer has timed out so flip LED output
       if (Alarm) digitalWrite(Buzzer, Buzzer_On);  // if enabled switch alarm buzzer on
       Loop_Timer = millis();              // Set LED flash timer for OFF cycle
       Check_Time_Control_Relay1();   // Timer Control for relay 1
       if (No_Of_Relays >= 2) Check_Time_Control_Relay2();   // Timer Control for relay 2 if fitted
       if (No_Of_Relays == 3) Check_Time_Control_Relay3();   // Timer Control for relay 3 if fitted          
                
     }  // end of      if (LoopTimer >= LED_On_Timer)      
   } // end of if (digitalRead(Heartbeat_LED))
   else              // If the Heartbeat LED is OFF
   {
    if (LoopTimer >= LED_Off_Timer) 
    {
      digitalWrite(Heartbeat_LED,LED_On);  // Off timer has timed out so flip LED output
      if (Alarm) digitalWrite(Buzzer, Buzzer_Off);  // if enabled, switch alarm buzzer off
      Process_Minute_Timers();  // check and update minute timer (re output timers and timer control)
      Check_Pump_Health();   // If enabled, check pump protection function (Relay 1 only)
      Interval_Update_Clock();  // Check soft clock against NTP server 
      Check_Timer_Reset_Clock(); // Check the Automatic Timer Reset clock  
      Loop_Timer = millis();              // Set LED flash timer for ON cycle
    }
   }  // end of else  // If the Heartbeat LED is OFF
} // end of void loop()
// -------------------------------------------------------- 
byte Check_AP_Switch() // returns value depending on how long the AP button is pressed  
{
  unsigned long Switch_Pressed_Time; 
  if(digitalRead(AP_Switch) == AP_Switch_Off )   // read & test the button status
  {       // if not pressed
      Time_Start = millis();   // get time marker in readiness for switch being pressed
      return 0;
  }
  else // switch is ON
  {
    while (digitalRead(AP_Switch) == AP_Switch_On)  // wait until button is released
    {
      if (AUX_Mode[1] == 2) Kill_Alarm();    // clear alarm if pump protection is ON and buzzer is running 
      Heartbeat_Tick();   // keep HW watchdog alive
      Switch_Pressed_Time = millis() - Time_Start;  // calculate the time since button was pressed(mS)
    }
          // button released
    if (Enable_Messages) Serial.println(Switch_Pressed_Time);     // Diagnostic - button duration time (mS) 
    if(Switch_Pressed_Time >= 10000) Reset_Device("by AP button");   // if AP_Switch is pressed for > 10 seconds = reset 
    if(Switch_Pressed_Time >= 5000)  return 2;   // Wake up Access point
    // if(Switch_Pressed_Time >= 2000) return 1;   // manual operation of relay output
    if(Switch_Pressed_Time <  5000)  return 0;   // button not pressed long enough

  }  // end of else // switch is ON
}     // end of byte Check_AP_Switch()
//  ---------------------------------------------------------------------
void Process_Minute_Timers()   // check the various function minute timers inc soft clock
{
  unsigned long Minute_Timer = millis() - Timer_Tick;
  if (Minute_Timer >= One_Minute_timer_Tick_Threshold)   // Check for 1 minute timer tick
  {     // every minute
    Timer_Tick = millis();
    
    if (WiFi.status() == !WL_CONNECTED)
    { 
        Lost_Connection_Timer ++;  // increment lost connection timer if no connection
        if(Lost_Connection_Timer >= Lost_Connection_Timeout) Reset_Device("Lost connection");
    }
    else Lost_Connection_Timer = Lost_Connection_Timeout;  //Connected so reset time out counter
    
    byte i = 0;   // check each ON timer for all activated outputs
    while (i < No_Of_Relays)
    {
      Check_Outputs_For_Timeout(i);  // check timer every minute, if time out switch output off
      i ++;
    }
    Clock_Minute ++;    // Service the soft clock
    if (Clock_Minute >= 60)  // if on the hour
    {
      Clock_Minute = 0;
      Clock_Hour ++;  // increment the hour value
      if (Clock_Hour >= 24) // if midnight
      {
        Clock_Hour = 0; 
        Clock_Week_Day ++; //increment week day
        if (Clock_Week_Day >= 8 || Clock_Week_Day == 0) Clock_Week_Day = 1;  // overflow from Saturday to Monday
      }
    }  
  }  // end of if (Minute_Timer >= One_Minute_timer_Tick_Threshold)
}
// ------------------------- Alexa voice driven commands -------------------------------
void SwitchOneOn() 
{
    if (Enable_Messages)
    {
      Serial.print(Relay_Descriptors[0]); 
      Serial.println(" turned on ...");
    }
    Activate_Relay(0, Switch_On,!Timer_Active[0]); // switch relay & virtual output flag to ON, 
                                                                   // only set output timer ON if relay timer is NOT on
}
// --------------------------------------------------------
void SwitchOneOff() 
{
    if (Enable_Messages)
    {
      Serial.print(Relay_Descriptors[0]); 
      Serial.println(" turned off ...");
    }
    Activate_Relay(0, Switch_Off,false);
}
// --------------------------------------------------------
void SwitchTwoOn() 
{
    if (Enable_Messages)
    {
      Serial.print(Relay_Descriptors[1]); 
      Serial.println(" turned on ...");
    }
    Activate_Relay(1, Switch_On,!Timer_Active[1]);
}
// --------------------------------------------------------
void SwitchTwoOff() 
{
    if (Enable_Messages)
    {
      Serial.print(Relay_Descriptors[1]); 
      Serial.println(" turned off ...");
    }
    Activate_Relay(1, Switch_Off,false);  
}
// --------------------------------------------------------
void SwitchThreeOn() 
{
    if (Enable_Messages)
    {
      Serial.print(Relay_Descriptors[2]); 
      Serial.println(" turned on ...");
    }
    Activate_Relay(2, Switch_On,!Timer_Active[2]);
}
// --------------------------------------------------------
void SwitchThreeOff() 
{
    if (Enable_Messages)
    {
      Serial.print(Relay_Descriptors[2]); 
      Serial.println(" turned off ...");
    }
    Activate_Relay(2, Switch_Off,false);    
}
// -------------------------------------------------------- 
void Check_For_Keyboard_Command()  // Used in loop() - checks keyboard for a single character command
{
  if (Serial.available() > 0) Command = Serial.read();
  {
    if (Command == 'd')  // Download and display all EEPROM based values
    {
      Download_Watchdog_Type_And_Alexa_Enable(true);       
      Enable_Messages = EEPROM.read(EEPROM_Enable_Messages);
      Serial.print(F("Diagnostic messages enable = ")); Serial.println(Enable_Messages);
      Download_WiFi_Login_Details_From_EEPROM(true);
      Download_From_EEPROM_Relay_Descriptors(true);        
      Download_From_EEPROM_Output_And_Input_Settings(true);  
      Download_From_EEPROM_Timer_Control_Parameters(true);
      Download_Timer_Reset_Clock_From_EEPROM(true);     
    }   // end of if (Command == 'd')  // Download and display all EEPROM based values
    if (Command == 'A')  Switch_Alexa_Enable_Voice_Control(true);  // Turn ON voice control
    if (Command == 'B')  Switch_Alexa_Enable_Voice_Control(false); // Turn OFF voice control  
    if (Command == 'h')  Help();
    if (Command == 'r')  Remote_Read_Of_Relay_Status();   // Send output status values to serial port
    if (Command == 'w')  Remote_Control_Of_Relay(); // Set/reset outputs from serial port
    if (Command == 'W')  Display_WiFi_Connection_Status();
    if (Command == 'R')  Reset_Device("by keyboard");  // Reset device
    if (Command == 'E')  Reset_All_EEPROM();
    if (Command == 't')  Display_Current_Output_Timer_Values(); 
    if (Command == 'T')  Print_Clock();  
    if (Command == 'l')  Setup_New_WiFi_Login_Values();
    if (Command == 'f')  Diagnostic_Flag_Dump();
    if (Command == 'm')  Toggle_Diagnostic_Messages(); 
    if (Command == 'n')  Setup_No_Of_Relays();             
    if (Command == 'p')  Setup_No_Of_AUX_Inputs();  
    if (Command == 'a')  Toggle_Soft_Access_Point_Switch(); 
    if (Command == 'i')  Toggle_Timer_Reset(); 
    if (Command == 'b')  Setup_AP_Password();
    if (Command == 'x')  
    {
      if(AUX_Mode[0] == 2)   // if pump protection is on
      {
        if (Alarm) Kill_Alarm();   // if alarm is On, stop it
        else  Alarm = true;        // if alarm is off, enable it
      }
    } // end of    if (Command == 'x')
    if (Command == 'y')  Toggle_WatchDog_Type();
          
    Command = 255;
  }   // end of if (Serial.available() > 0) Command = Serial.read();
}     // end of void Check_For_Keyboard_Command() 
//  ---------------------------------------------------------------------
void Download_Watchdog_Type_And_Alexa_Enable(boolean Display)
{
  WatchDog_Type = EEPROM.read(EEPROM_WatchDog_Type);   // set working memory  
  Alexa_Enable[0] = EEPROM.read(EEPROM_Alexa_Enable1);
  Alexa_Enable[1] = EEPROM.read(EEPROM_Alexa_Enable2);
  Alexa_Enable[2] = EEPROM.read(EEPROM_Alexa_Enable3);  
  if(Display) 
  {
    Serial.println("");
    Serial.print(F("Watchdog type = ")); 
    if (WatchDog_Type == Hardware) Serial.println(F("Hardware")); else Serial.println(F("Software")); 
    Serial.print(F("Amazon Alexa Output 1 voice control is  "));    
    if (Alexa_Enable[0]) Serial.println(F("enabled")); else Serial.println(F("disabled"));
    Serial.print(F("Amazon Alexa Output 2 voice control is  "));    
    if (Alexa_Enable[1]) Serial.println(F("enabled")); else Serial.println(F("disabled"));
    Serial.print(F("Amazon Alexa Output 3 voice control is  "));    
    if (Alexa_Enable[2]) Serial.println(F("enabled")); else Serial.println(F("disabled"));
  } 
}
//  ---------------------------------------------------------------------
void Setup_No_Of_Relays()
{
  Serial.println(""); Serial.println(F("How many relays fitted (1-3)? <You must reset the device before the new setting is loaded>"));
  byte Length = GetKeyboardInput();   // Get keyboard entry
  byte value = Validate_Numerical_Entry_From_Keyboard_Buffer(Length, 3);  // validate it for value 1 to 3
  if (value == 255)
  {
    Serial.println(F("Invalid Entry"));
    return; // not valid
  }
  if (value == 0) value = 1;
  Serial.print(F("No of relays fitted = ")); Serial.println(value);
  EEPROM.write(EEPROM_No_Of_Relays, value);
  EEPROM.commit();
}
//  ---------------------------------------------------------------------
void Setup_No_Of_AUX_Inputs()
{
  Serial.println(""); Serial.println(F("How many AUX inputs are fitted to the board (0-3) "));
  byte Length = GetKeyboardInput();   // Get keyboard entry
  byte value = Validate_Numerical_Entry_From_Keyboard_Buffer(Length, 3);  // validate it for value 0 to 3
  if (value == 255)
  {
    Serial.println(F("Invalid Entry"));
    return; // not valid
  }
  Serial.print(F("No of Auxiliary inputs fitted to the board = ")); Serial.println(value);
  EEPROM.write(EEPROM_No_Of_AUX, value);
  EEPROM.commit();
  No_Of_AUX = value; 
}
  // --------------------------------------------------------------------
byte Validate_Numerical_Entry_From_Keyboard_Buffer(byte No_Of_Keystrokes, byte Value_Limit) // validate characters in the keyboard buffer and return a value
{
  byte Value;   // numerical value to be returned from the keyboard buffer.  255 = invalid entry
  boolean Invalid_Entry = false;
  // test depending on number of keystrokes
  if (No_Of_Keystrokes == 1) if(isDigit (KeyboardBuffer[0])) Value = KeyboardBuffer[0] - '0'; else Invalid_Entry = true;      
  if (No_Of_Keystrokes == 2)
  {
    if (isDigit(KeyboardBuffer[1]) && isDigit(KeyboardBuffer[0])) 
      Value = (KeyboardBuffer[1] - '0') + ((KeyboardBuffer[0] - '0') * 10); else Invalid_Entry = true;
  }
  if (No_Of_Keystrokes == 3) 
  {
    if (isDigit(KeyboardBuffer[2]) && isDigit(KeyboardBuffer[1]) && isDigit(KeyboardBuffer[0])) 
      Value = (KeyboardBuffer[2] - '0') + ((KeyboardBuffer[1] - '0') * 10) + ((KeyboardBuffer[0] - '0') * 100) ;
    else Invalid_Entry = true;      // if not numeric  
  }             
       // test for validity - range, null entry, too many characters
  if (No_Of_Keystrokes > 3 || No_Of_Keystrokes == 0 || Value > Value_Limit) Invalid_Entry = true;
  if (Invalid_Entry)
  {
    Serial.println(""); Serial.println(F("Invalid Entry")); Serial.println("");
    Value = 255;
    return Value;   // illegal data entry      
  }
  else return Value;   // value = numerical keyboard entry or = 255 if not valid  
}
    // ------------------------------------------------------------------  
void Toggle_Diagnostic_Messages()  // enable/disable display printout of messages
{
  Enable_Messages = !EEPROM.read(EEPROM_Enable_Messages);  // get and invert the current flag status
  EEPROM.write(EEPROM_Enable_Messages, Enable_Messages);         // re-write the inverted flag
  EEPROM.commit();
  Serial.print(F("Diagnostic messages are now "));
  if (Enable_Messages) Serial.println(F("enabled"));
  else Serial.println(F("Diagnostic messages are now disabled"));
}  // end of void Toggle_Diagnostic_Messages()
//  ---------------------------------------------------------------------
void Toggle_WatchDog_Type()  // toggle between Hardware/Software watchdog type
{
  WatchDog_Type = !EEPROM.read(EEPROM_WatchDog_Type);  // get and invert the current flag status
  EEPROM.write(EEPROM_WatchDog_Type, WatchDog_Type);         // re-write the inverted flag
  EEPROM.commit();
  Serial.print(F("Watchdog is now "));
  if (WatchDog_Type) Serial.println(F("hardware based"));
  else Serial.println(F("software based"));
}  // end of void Toggle_WatchDog_Type()
//  ---------------------------------------------------------------------
void Toggle_Timer_Reset()  // enable/disable of Automatic Timer Reset
{
  Timer_Reset_Enable = !Timer_Reset_Enable;  // invert the Flag
  Serial.print(F("Automatic Timer Reset is "));
  if (Timer_Reset_Enable) Serial.println(F("enabled "));
  else Serial.println(F("disabled"));
}  // end of void Toggle_Timer_Reset()  // enable/disable of Automatic Timer Reset
//  ---------------------------------------------------------------------
void Toggle_Soft_Access_Point_Switch()  // enable/disable Access point
{
  Soft_AP_Switch = !Soft_AP_Switch;                              // write to working memory
  Serial.print(F("Access Point is "));
  if (Soft_AP_Switch) Serial.println(F("enabled"));
  else Serial.println(F("disabled"));
}  // end of Toggle_Soft_Access_Point_Switch()
//  ---------------------------------------------------------------------
void Switch_Alexa_Enable_Voice_Control(boolean On_Off)   // true = enable
{  
  Serial.print(F("Alexa voice control is "));
  if (On_Off) Serial.print(F("enabled"));
  else  Serial.print(F("disabled"));
  Serial.print(F(", device will now reset"));
  EEPROM.write(EEPROM_Alexa_Enable1,On_Off);    // write to EEPROM memory
  EEPROM.write(EEPROM_Alexa_Enable2,On_Off);    // write to EEPROM memory
  EEPROM.write(EEPROM_Alexa_Enable3,On_Off);    // write to EEPROM memory    
  EEPROM.commit();  
  Reset_Device("Alexa Enable changed");   
}
//  ---------------------------------------------------------------------
void Diagnostic_Flag_Dump()
{
  Serial.println("");
  Serial.println(F("DEMON Status Dump"));
  Serial.print(F("No of Relays = ")); Serial.println(No_Of_Relays); 
  Serial.print(F("No of AUX Inputs = ")); Serial.println(No_Of_AUX);   
  Serial.print(F("Watchdog is "));
  if (WatchDog_Type == Hardware) Serial.println(F("Hardware based")); else Serial.println(F("Software based"));
  Serial.println("");
  Serial.print(F("Alexa Output 1 Voice Control is "));
  if (Alexa_Enable[0]) Serial.println(F("enabled")); else Serial.println(F("disabled"));  
  Serial.print(F("Alexa Output 2 Voice Control is "));
  if (Alexa_Enable[1]) Serial.println(F("enabled")); else Serial.println(F("disabled"));
  Serial.print(F("Alexa Output 3 Voice Control is "));
  if (Alexa_Enable[2]) Serial.println(F("enabled")); else Serial.println(F("disabled")); 
  Serial.println("");  
  Serial.print(F("The Setup Access Point is "));
  if(Access_Point_Running) Serial.println(F("running")); else  Serial.println(F("not running")); 
  Serial.print(F("The Soft Access Point Switch is ")); 
  if(Soft_AP_Switch) Serial.println(F("ON")); else  Serial.println(F("OFF"));  
  Serial.println("");  
  Serial.print(F("Initial login was ")); 
  if(Login_Successful) Serial.println(F("successful")); else Serial.println(F("not successful"));
  Serial.print(F("The NTP_Time_Interval (mS) is ")); Serial.println(NTP_Time_Interval);
  Serial.print(F("The last NTP Clock Update ")); 
  if(NTP_Resolved) Serial.println(F("was successful"));  else Serial.println(F("has failed"));
  Serial.print(F("Automatic Timer Reset is ")); 
  if(Timer_Reset_Enable) Serial.println(F("Enabled")); else  Serial.println(F("Disabled"));
  Serial.print(F("Serial port diagnostic messaging is ")); 
  if(Enable_Messages) Serial.println(F("Enabled")); else  Serial.println(F("Disabled")); 
  Serial.println("");   
  Serial.print(F("Relay 1 Timer Control is ")); 
  if(Timer_Active[0]) Serial.println(F("Running")); else  Serial.println(F("Off"));
  if (No_Of_Relays >= 2) 
  {
    Serial.print(F("Relay 2 Timer Control is ")); 
    if(Timer_Active[1]) Serial.println(F("Running")); else  Serial.println(F("Off"));
  } // end of if (No_Of_Relays >= 2) 
  if (No_Of_Relays == 3)
  {
    Serial.print(F("Relay 3 Timer Control is ")); 
    if(Timer_Active[2]) Serial.println(F("Running")); else  Serial.println(F("Off"));  
  }    // end of if (No_Of_Relays == 3)
}   // end of  void Diagnostic_Flag_Dump()
//  ---------------------------------------------------------------------
void Remote_Read_Of_Relay_Status()   // displays the current state of the fitted relays
{   
  byte i = 0;
  boolean state;
  while (i < No_Of_Relays)
  {
    state = digitalRead(Relay_Pin[i]);
    if (state == Relay_On) state = true; else state = false;   
    Serial.println(state);
    i++;  
  }  
}   // end of void Remote_Read_Of_Relay_Status()
//  ---------------------------------------------------------------------
void Remote_Control_Of_Relay()  // Remote control of any active output. w11 = set out 1, w10 = reset output 1,  
{                               // w21 = set output 2, w20 = reset output 2 etc
  byte Output_No;
  byte State = 0;  
  byte i = 0;  
  delay(2);
  while (Serial.available()>0) 
  {
    delay(2);
    char ch = Serial.read();
    KeyboardBuffer[i] = ch;   // get data sent after the command letter
    i++;        
  }   // end of while (Serial.available()>0)
  if (isDigit(KeyboardBuffer[1]) && isDigit(KeyboardBuffer[0])) // check 2 characters are numeric
  {
    Output_No = ((KeyboardBuffer[0] - '0')); // first char = output number
    State = (KeyboardBuffer[1] - '0');       // 2nd character = output status
    if(State) State = Relay_On; else State = Relay_Off;
    Activate_Relay(Output_No-1,State,!Timer_Active[Output_No-1]);
  }  // end of if (isDigit(KeyboardBuffer[1]) && isDigit(KeyboardBuffer[0]))
}    // end of void Remote_Control_Of_Relay()
//  ---------------------------------------------------------------------
void Help()  // displays all of the keyboard commands
{
    Serial.println(""); Serial.println(Firmware_Version);Serial.println("");
    Serial.println(F("AVAILABLE KEYBOARD COMMANDS")); Serial.println("");
    Serial.println(F("a = Toggle the soft Access Point Enable Flag")); 
    Serial.println(F("A = Turn ON voice control for all outputs"));
    Serial.println(F("B = Turn OFF voice control for all outputs")); 
    Serial.println(F("b = Set up Configuration Access Point password"));        
    Serial.println(F("d = Download and display all of the EEPROM stored configuration parameters"));
    Serial.println(F("E = Clear all EEPROM to default values (Must reset afterward)"));
    Serial.println(F("f = Diagnostic Demon - Active Flags & Variables")); 
    Serial.println(F("h = Show this Help menu screen again"));       
    Serial.println(F("i = Toggle the Automatic Timer Reset enable Flag"));           
    Serial.println(F("l = Keyboard entry of WiFi login parameters"));
    Serial.println(F("m = Toggle Enable Diagnostic Messsages"));
    Serial.println(F("n = Set the number of relays outputs <Must reset afterwards>")); 
    Serial.println(F("p = Set the number of AUX inputs")); 
    Serial.println(F("R = Reset device"));
    Serial.println(F("t = Display current output timer values"));
    Serial.println(F("T = Display Soft Clock"));        
    Serial.println(F("w = Direct write to output E.g w10, w11, w20, w21, w30, w31")); 
    Serial.println(F("W = Display WiFi connection status and details"));      
    Serial.println(F("x = Toggle the Alarm Enable Flag (Only if pump protection is on)"));  
    Serial.println(F("y = Toggle the watchdog type between hardware and software base"));        
    Serial.println("");
}  // end of void Help()

//  ---------------------------------------------------------------------
void Check_Outputs_For_Timeout(byte Output_No)    // every minute check outputs for possible time out (output timers)
{
  if((digitalRead(Relay_Pin[Output_No]) == Relay_Off) || Relay_On_Timers[Output_No] == 0) return;  // nothing to do if relay is OFF or timer disabled
  Relay_On_Timers[Output_No] = Relay_On_Timers[Output_No] - 1;   // decrement timer
  if (Relay_On_Timers[Output_No] > 0) return;  // nothing to do if timer has not timed out
  Activate_Relay(Output_No,Relay_Off,true);  // Switch relay and remote control flag OFF
}  // end of void Check_Outputs_For_Timeout(byte Output_No)
// ----------------------------------------------------------------------
void Display_Current_Output_Timer_Values()  // displays the current timer values for each fitted relay
{
  byte i = 0;
  while(i < No_Of_Relays)  // For each output, display timer value
  {
    Serial.println("");
    Serial.print(F("Output timer ")); 
    Serial.print(i+1); 
    Serial.print(F(" = ")); 
    Serial.print(Relay_On_Timers[i]);
    Serial.println(F(" minutes"));
    i ++;
  }
}  // end of void Display_Current_Output_Timer_Values()
//  ---------------------------------------------------------------------
void Check_PIR_Inputs(byte PIR_No)  // check for PIR activity and switch lamps on if enabled or set alarm flag
{
  if (AUX_Mode[PIR_No] == 1) // only if Input is enabled as a PIR sensor
  {
    if (digitalRead(AUX_Pin[PIR_No]) == PIR_Off)  // nothing to do if PIR is enabled but not ON
    {
      PIR_Switch_Flag[PIR_No] = false;  // ensure that PIR flag is reset
      return;
    }
    if((digitalRead(Relay_Pin[PIR_No]) == Relay_Off))  // PIR is ON and enabled, check if the relay output is Off
    {   // relay is off therefore need to turn it on
      Activate_Relay(PIR_No,Relay_On,false);    // if enabled, switch relay and flag ON 

      PIR_Switch_Flag[PIR_No] = true;  // set PIR flag to indicate that it was the PIR that switched the output on
      Relay_On_Timers[PIR_No] = PIR_Time; // set output timer to PIR time value
      return;
    }   // end of   if((digitalRead(Relay_Pin[PIR_No]) == Relay_Off))
    else // PIR is on and enabled & the relay is already on
    {
      if (PIR_Switch_Flag[PIR_No] == false) return;  // nothing to do since it was not the PIR that initiated the switch on
      else Relay_On_Timers[PIR_No] = PIR_Time;    // if enabled, keep the timer set to the PIR value
      return;
    }   // end of else // PIR is on and enabled & the relay is already on
  } // end of   if (AUX_Mode[PIR_No] == 1) // only if Input is enabled as a PIR sensor
  else return;   // Input not enabled as a PIR so exit immediately
}  // end of void Check_PIR_Inputs(byte PIR_No)
// ------------------------------------------------------------
void Check_AUX_Button()   // check AUX inputs are not functioning as manual on/off buttons
{                         // for the relays 
  byte i = 0;
  boolean State;
  while (i < No_Of_AUX)   // for each enabled AUX input
  {
    if (AUX_Mode[i] == 3 && digitalRead(AUX_Pin[i]) == Switch_On)   // only if AUX input set as a switch & button pressed
    {
      while (digitalRead(AUX_Pin[i]) == Switch_On)   // wait while the button is pressed
      {
        Heartbeat_Tick();   // keep HW watchdog alive
      }
      State = digitalRead(Relay_Pin[i]);  // read relay state after button is released
      Activate_Relay(i, !State, 0);     // toggle the relay state
    }   // end of if (AUX_Mode[i] == 3 && digitalRead(AUX_Pin[i]) == LOW)
    i ++;
  } // end of while (i < No_Of_AUX)   // for each enabled AUX input
}   // end of void Check_AUX_Button()
// ------------------------------------------------------------
void Reset_All_EEPROM()   // Clear out all of the previous stored configuration entries & reset the device
{                         // to default values
  unsigned int i = EEPROM_Start_Address;   // initial clear out of complete EEPROM block
  while(i < 300)  
  {
     EEPROM.write(EEPROM_Start_Address + i , 255);
     i++;
  }
  i = EEPROM_WiFi_SSID;
  while(i < EEPROM_WiFi_SSID + Max_String_Length)  // stuff default name into SSID
  {
    EEPROM.write(i, 'S');
    i++;
  }
  i = EEPROM_WiFi_Password;
  while(i < (EEPROM_WiFi_Password + Max_String_Length))  // stuff default name into Password
  {
    EEPROM.write(i, 'P'); 
    i++;
  }
  i = EEPROM_HostName;
  while(i < (EEPROM_HostName + Max_String_Length))  // stuff default name into HostName
  {
    EEPROM.write(i, 'N'); 
    i++;
  }
  EEPROM.write (EEPROM_AUX_Mode, 0);   // operating mode of AUX1 input = disabled
  EEPROM.write (EEPROM_AUX_Mode + 1, 0); // operating mode of AUX2 input = disabled 
  EEPROM.write (EEPROM_AUX_Mode + 2, 0); // operating mode of AUX3 input = disabled   
  EEPROM.write (EEPROM_PIR_Time, 10);  // PIR ON time = 10 minutes
  EEPROM.write (EEPROM_Relay_On_Time_Parameter,60); // Output 1 timer = 60
  EEPROM.write (EEPROM_Relay_On_Time_Parameter + 1,60); // Output 2 timer = 60
  EEPROM.write (EEPROM_Relay_On_Time_Parameter + 2,60); // Output 3 timer = 60
  EEPROM.write (EEPROM_Timer_Reset_Clock_Mins,30);
  EEPROM.write (EEPROM_Timer_Reset_Clock_Hrs,7);
  EEPROM.write (EEPROM_Timer_Reset_Enable,false);  // Alexa/NTP master clock = disabled  
  EEPROM.write (EEPROM_Enable_Messages,true);  // enable disgnostic messages
  EEPROM.write (EEPROM_No_Of_Relays,1);  // default no of relays
  EEPROM.write (EEPROM_No_Of_AUX,0);  // default no of AUX inputs 
  EEPROM.write (EEPROM_Timer_Enable_Relay1,false);   // Relay Time control disabled
  EEPROM.write (EEPROM_Timer_Enable_Relay2,false);
  EEPROM.write (EEPROM_Timer_Enable_Relay3,false);
  EEPROM.write (EEPROM_Pump_Protection_Start_Up_Delay_Time,20);
  EEPROM.write (EEPROM_Pump_Protection_Dry_Time,5);
  EEPROM.write (EEPROM_WatchDog_Type,Software);   // watchdog = software based
  EEPROM.write (EEPROM_Alexa_Enable1,false);   // Alexa disabled
  EEPROM.write (EEPROM_Alexa_Enable2,false);   // Alexa disabled
  EEPROM.write (EEPROM_Alexa_Enable3,false);   // Alexa disabled  
  EEPROM.write (EEPROM_AP_Password,'p');      // AP default password = password
  EEPROM.write (EEPROM_AP_Password+1,'a'); 
  EEPROM.write (EEPROM_AP_Password+2,'s'); 
  EEPROM.write (EEPROM_AP_Password+3,'s'); 
  EEPROM.write (EEPROM_AP_Password+4,'w'); 
  EEPROM.write (EEPROM_AP_Password+5,'o'); 
  EEPROM.write (EEPROM_AP_Password+6,'r'); 
  EEPROM.write (EEPROM_AP_Password+7,'d'); 
  i=8;
  while (i < Max_String_Length)
  {
    EEPROM.write (EEPROM_AP_Password+i,255); 
    i++;
  }
  
  i = 0;
  while (i < 45)    // reset boolean flags for each of 6 x timer week day flags and timer enable flags
  {
    EEPROM.write(EEPROM_Relay1_Timer1_Sun + i, false);
    i++;
  }
  i = 0;
  while (i < 12)    // clear all Timer ON & OFF control parameters
  {
    EEPROM.write(EEPROM_Relay1_Timer1_TimOn_Min + (2 * i),0);
    EEPROM.write(EEPROM_Relay1_Timer1_TimOn_Hrs + (2 * i),i);   
    i= i + 1;   
  } 

  byte Value;         // set up EEPROM header block - indicates that the EEPROM
  i = 0;              // has been initialised
  while (i < 5)
  {
    EEPROM.write(EEPROM_Header_Block + i, i);  // write header byte
    i ++;  
  }  // end of while (i < 5) 
     
  EEPROM.commit();
  Serial.println(F("EEPROM reset to default values"));
}   // end of void Reset_All_EEPROM()
// ------------------------------------------------------------
void Download_WiFi_Login_Details_From_EEPROM(boolean Display)   // Download to memory, the EEPROM stored values of SSID and password
{
    unsigned int i = EEPROM_WiFi_SSID;
    esid = "";  //start with empty SSID string
    while (i < EEPROM_WiFi_SSID + Max_String_Length)  // build up SSID string from stored characters that are not FF
    {
      if (EEPROM.read(i) != 255) 
      {
      esid += char (EEPROM.read(i)); 
      } else break;
      i++; 
    }
    i = EEPROM_WiFi_Password;
    epass = "";   //start with empty password string
    while (i < EEPROM_WiFi_Password + Max_String_Length) // build up password string from stored characters that are not FF
    {
      if (EEPROM.read(i) != 255) 
      {
      epass += char(EEPROM.read(i)); 
      } else break; 
      i++;
    }
    i = EEPROM_HostName;
    HostName = "";   //start with empty password string
    while (i < EEPROM_HostName + Max_String_Length) // build up password string from stored characters that are not FF
    {
      if (EEPROM.read(i) != 255) 
      {
      HostName += char(EEPROM.read(i)); 
      } else break; 
      i++;
    }

    i = EEPROM_AP_Password;
    passphrase = "";   //start with empty password string
    while (i < EEPROM_AP_Password + Max_String_Length) // build up password string from stored characters that are not FF
    {
      if (EEPROM.read(i) != 255) 
      {
      passphrase += char(EEPROM.read(i)); 
      } else break; 
      i++;
    }
    if (Display)
    {
      Serial.println(""); 
      Serial.print(F("EEPROM SSID: ")); 
      Serial.println(esid); // print final SSID string - diagnostic
      // Serial.print(F("EEPROM Password: ")); 
      // Serial.println(epass);  // print final Password string - diagnostic
      Serial.print(F("EEPROM HostName: ")); 
      Serial.println(HostName);  // print final HostName string - diagnostic
      Serial.print(F("EEPROM Access Point Password: ")); 
      Serial.println(passphrase);     
      Serial.println("");
    }    
}   // end of void Download_WiFi_Login_Details_From_EEPROM()
// ------------------------------------------------------------
void Display_WiFi_Connection_Status()  // display the status of the current WiFi connection
{
    Serial.println(""); 
               // print out to serial monitor, the connected network details 
    Serial.print(F("Device name: "));
    Serial.println(HostName);

    Serial.print(F("SSID: "));      // SSID
    Serial.println(WiFi.SSID());                       
  
    long rssi = WiFi.RSSI();        // print the received signal strength:
    Serial.print(F("signal strength (RSSI):"));
    Serial.println(rssi);            
  
    IPAddress ip = WiFi.localIP();
    Serial.print(F("IP Address: ")); // IP address given to the connection
    Serial.println(ip);
  
    byte mac[6];                      // MAC address 
    WiFi.macAddress(mac);
    Serial.print(F("MAC address: "));
    Serial.print(mac[5],HEX);
    Serial.print(F(":"));
    Serial.print(mac[4],HEX);
    Serial.print(F(":"));
    Serial.print(mac[3],HEX);
    Serial.print(F(":"));
    Serial.print(mac[2],HEX);
    Serial.print(F(":"));
    Serial.print(mac[1],HEX);
    Serial.print(F(":"));
    Serial.println(mac[0],HEX);

    IPAddress subnet = WiFi.subnetMask();   // Subnet mask
    Serial.print(F("NetMask: "));
    Serial.println(subnet);
  
    IPAddress gateway = WiFi.gatewayIP();   // Gateway address
    Serial.print(F("Gateway: "));
    Serial.println(gateway);
    Serial.println("");
  }  // end of void Display_WiFi_Connection_status

// ---------------------- WIFI --------------------------------------
void WifiConnect()     // Connect to the wifi network using the stored credentials
{        
   WiFi.mode(WIFI_STA);             // Set WIFI module to STA mode
   byte Login_Wait_Timer = 0;    // no of loops waiting for log in success
   Login_Successful = true;
   if (Enable_Messages) 
   {  
      Serial.println(); 
      Serial.println(F("Connecting to Local WiFi Network"));
   }
   WiFi.begin(esid.c_str(), epass.c_str());  
   delay(2);
   WiFi.hostname(HostName.c_str());  
   while (WiFi.status() != WL_CONNECTED) 
   {                 // wait for connection whilst flashing the heartbeat LED
     Check_For_Keyboard_Command();  // check for any keyboard command     
     digitalWrite(Heartbeat_LED,LED_On);    // flash the Heartbeat LED
     delay(LED_On_Timer);
     if (Enable_Messages) Serial.print(".");
     digitalWrite(Heartbeat_LED,LED_Off);
     delay(LED_Off_Timer );   
     Login_Wait_Timer ++;
     if ( Login_Wait_Timer >= 200)  // login failed
     {
        Login_Successful = false;
        Soft_AP_Switch = true;      // enable Access point to allow set up check
        break;
     }         
   }  // end of while (WiFi.status() != WL_CONNECTED) 
   if(Login_Successful) 
   {
      if (Enable_Messages) 
      {  
          Serial.println("");
          Serial.println(F(" Connected to Network"));
      } 
    }  // end of if(Login_Successful)
   else
   {   
      if (Enable_Messages)
      {
          Serial.println(); 
          Serial.println(F("Login failed"));
      }
      LED_On_Timer = LED_On_Flash ;    // Connected to network so reset the LED flash timers for the main loop
      LED_Off_Timer = LED_Off_Flash; 
   } // end of else
 } // end of void WifiConnect()
// ----------------------- NTP functions ---------------------------------------
// send an NTP request to the time server at the given address
unsigned long sendNTPpacket(IPAddress& address)
{       // set all bytes in the buffer to 0
  memset(packetBuffer2, 0, NTP_PACKET_SIZE);
            // Initialize values needed to form NTP request
  packetBuffer2[0] = 0b11100011;   // LI, Version, Mode
  packetBuffer2[1] = 0;     // Stratum, or type of clock
  packetBuffer2[2] = 6;     // Polling Interval
  packetBuffer2[3] = 0xEC;  // Peer Clock Precision
                            // 8 bytes of zero for Root Delay & Root Dispersion
  packetBuffer2[12]  = 49;
  packetBuffer2[13]  = 0x4E;
  packetBuffer2[14]  = 49;
  packetBuffer2[15]  = 52;
                            // you can send a packet requesting a timestamp:
  udp.beginPacket(address, 123); //NTP requests are to port 123
  udp.write(packetBuffer2, NTP_PACKET_SIZE);
  udp.endPacket();
} // end of unsigned long sendNTPpacket(IPAddress& address)
    // ---------------------------------------------------
void Interval_Update_Clock() // update the soft clock with the NTP clock dataT
{
   unsigned long NTPTimer = millis() - NTP_Timer;
   if (NTPTimer >= NTP_Time_Interval)   // only do this at defined intervals
   {
     NTP_Timer = millis();
     Get_Clock_Data();   // get latest time via NTP and if valid, update soft clock
   }
}   //end of  Interval_Update_Clock()
    // ---------------------------------------------------
void Get_Clock_Data() // update the soft clock with the NTP clock data
{                        // Returns false if NTP fails
    unsigned long Raw_Time = Get_NTP_Time();  // get the NTP time
    if ( Raw_Time > 1512259200L) // check for validity against a random     
    {                            // already passed date - 12AM 3rd December                      
      Clock_Hour = byte(((Raw_Time % 86400L) / 3600) + Time_Zone);   // Update the soft clock with NTP data
      Clock_Minute = byte((Raw_Time % 3600) / 60);
      Clock_Week_Day = byte(((Raw_Time % 604800L) / 86400)); 
      if (Clock_Week_Day < 3) Clock_Week_Day = Clock_Week_Day + 5;
      else  Clock_Week_Day = Clock_Week_Day - 2;      // correct for week day offset
      NTP_Time_Interval = 10680000; // set next timed update for 2 hour, 58 minutes 
      NTP_Resolved = true;    
    }
    else    // clock data was invalid so set reduced interval timer for
    {       // another attempt at NTP
      NTP_Time_Interval = 10000;  
      NTP_Resolved = false;
      if (Enable_Messages) 
      {
          Serial.println(""); 
          Serial.println("NTP Failed"); 
      }
     }
    if (Enable_Messages) 
    {
            // print the hour, minute   
      Serial.println("");
      Serial.print(F("NTP Time is ")); Serial.print(Clock_Hour); // print the hour
      Serial.print(':');
      if (Clock_Minute < 10 ) Serial.print('0');
      Serial.println(Clock_Minute); // print the minute 
      Serial.print(F("Week day is ")); Serial.println(Clock_Week_Day);   // print the day of the week  Sunday = 1    
      Serial.println("");
    }
}   // end of Get_Clock_data
    // ---------------------------------------------------
time_t Get_NTP_Time()   // request timestamp from NTP server. returns 0 if fail
{
    while (udp.parsePacket() > 0) ; // discard any previously received packets
    
    udp.begin(localPort);
    
    // WiFi.hostByName(ntpServerName, timeServerIP); 
    // sendNTPpacket(timeServerIP); // send an NTP packet to a time server
    
    sendNTPpacket(timeServer); // send an NTP packet to a time server
    
    // wait to see if a reply is available
    byte NTP_Wait_Timer = 0;
    boolean cb = false;
    unsigned long epoch = 0;
    while (NTP_Wait_Timer < NTP_Wait_Time && !cb)  // wait upto 1 second for NTP server to respond
    {
      digitalWrite(Heartbeat_LED,LED_On);  // keep HW watchdog alive
      delay(5);
      cb = udp.parsePacket();
      NTP_Wait_Timer ++;
      digitalWrite(Heartbeat_LED,LED_Off);
      NTP_Timeout = false;       
      delay(5);
    }
    if (NTP_Wait_Timer >= NTP_Wait_Time) // check to see if wait timer timed out
    {         // time out - attempt failed so set timer update interval to 15 seconds for next attempt
      NTP_Timeout = true; 
      return epoch; // nothing received back
    }
    else // data was received so read data
    {
      udp.read(packetBuffer2, NTP_PACKET_SIZE); // read the packet into the buffer
      unsigned long highWord = word(packetBuffer2[40], packetBuffer2[41]);
      unsigned long lowWord = word(packetBuffer2[42], packetBuffer2[43]);
      unsigned long secsSince1900 = highWord << 16 | lowWord;
      const unsigned long seventyYears = 2208988800UL;
      epoch = secsSince1900 - seventyYears;    // UTC time seconds
      return epoch;
    }  // end of else
   
   udp.stop();
   
}   // end of void Get_NTP_Time()
// ------------------------------------------------------------
byte GetKeyboardInput()    // get data string from keyboard - max 20 characters, terminator = carriage return
{                                           
   byte pointer = 0;      // flush keyboard buffer
   while(pointer < 20)
   {
    KeyboardBuffer[pointer] = 255;
    pointer ++;
   }
   while (Serial.available() > 0) {byte Trash = Serial.read();}   // flush out any trash data
   while (Serial.available() == false) Heartbeat_Tick();  // pause for keyboard input & flash 
                                                          // the Heartbeat LED to keep the watchdog alive   
   char ch;
   pointer = 0;
   while (Serial.available()>0)   //For each character entered
   {
     ch = Serial.read();
     if (pointer < Max_String_Length && byte(ch) != 13)    
     {
       KeyboardBuffer[pointer] = ch;  // save to keyboard buffer
       pointer ++; 
     }
     Heartbeat_Tick();   // flash the Heartbeat LED to keep the watchdog alive  
   }  // end of    while (Serial.available()>0)   //For each character entered
   return pointer;
}  // end of void GetKeyboardInput
// ------------------------------------------------------------
void Setup_New_WiFi_Login_Values()  // Set up new WiFi login - SSID, Password & Host Name
{
  Serial.println(""); Serial.println(F("Enter new SSID - 20 chars max"));
   
  byte Length = GetKeyboardInput() ;   // get input from keyboard, length = number of characters used 
  unsigned int i=EEPROM_WiFi_SSID;
  if (Length >= 8 && Length < 21)  // if entry is > minimum & less than the max no of characters 
  {
    while(i < EEPROM_WiFi_SSID + Max_String_Length)  // Clear out all of the previous stored entriesfor SSID
    {
      EEPROM.write(i, 255);
      i++;
    }
    Serial.print(F("New WiFi SSID = "));
    i=0;
    while(i < Length)   // store new SSID to EEPROM
    {
      EEPROM.write(EEPROM_WiFi_SSID + i, KeyboardBuffer[i]);
      Serial.print(char(EEPROM.read(EEPROM_WiFi_SSID + i))); 
      i++;
    }
  } // end of if (Length >= 8 && Length < 21)  // if entry is > minimum & less than the max no of characters
  else
  Serial.println(F("Invalid - WiFi SSID not changed"));   
  Serial.println("");
    
  Serial.println(F("Enter WiFi Password - 8 chars min, 20 chars max"));
  Length = GetKeyboardInput() ;   // get input from keyboard & how many characters entered   
  if(Length >= 8 && Length < 21)  // if entry is > minimum & less than the max no of characters 
  {   
    i = EEPROM_WiFi_Password;
    while(i < (EEPROM_WiFi_Password + Max_String_Length))
    {
      EEPROM.write(i, 255); // Clear out all of the previous stored entries for password
      i++;
    }
    Serial.print(F("WiFi Password = "));     
    i = 0; 
    while(i < Length)  // store new password to EEPROM
    {
      EEPROM.write(EEPROM_WiFi_Password + i, KeyboardBuffer[i]);
      Serial.print(char(EEPROM.read(EEPROM_WiFi_Password + i))); 
      i++;
    }
  } // end of if(Length >= 8 && Length < 21)  // if entry is > minimum & less than the max no of characters
  else
  Serial.println(F("Invalid - WiFi password not changed"));
  Serial.println("");
  
  Serial.println(F("Enter Device Name - 4 chars min, 20 chars max"));
  Length = GetKeyboardInput() ;   // get input from keyboard  & how many characters entered      
  i = EEPROM_HostName;
  if (Length >= 4 && Length  < 21)  // if entry is > minimum & less than the max no of characters
  {
    while(i < (EEPROM_HostName + Max_String_Length))
    {
      EEPROM.write(i, 255); // Clear out all of the previous stored entries
      i++;
    }
    Serial.print(F("Device Name = "));     
    i = 0; 
    while(i < Length)  // store new password to EEPROM
    {
      EEPROM.write(EEPROM_HostName + i, KeyboardBuffer[i]);
      Serial.print(char(EEPROM.read(EEPROM_HostName + i))); // ------------------
      i++;
    }
    Serial.println("");
  }   // end of  if (Length >= 4)  // if entry is > minimum no of characters
  else
  Serial.println(F("Invalid - Device Name not changed"));
  EEPROM.commit();   // update EEPROM   
  Download_WiFi_Login_Details_From_EEPROM(true); 
}  // end of  void Setup_New_WiFi_Login_Values()  
 // ---------------------------------------------------------------------------
void Setup_AP_Password()
{  
  Serial.println(F("Enter Configuration Access Point Password - 8 chars min, 20 chars max"));
  byte Length = GetKeyboardInput() ;   // get input from keyboard & how many characters entered   
  byte i = EEPROM_AP_Password;
  if (Length >= 8 && Length < 21)  // if entry is > minimum & less than the max no of characters
  {
    while(i < (EEPROM_AP_Password + Max_String_Length))
    {
      EEPROM.write(i, 255); // Clear out all of the previous stored entries
      i++;
    }
    Serial.print(F("New Access Point Password = "));     
    i = 0; 
    while(i < Length)  // store new password to EEPROM
    {
      EEPROM.write(EEPROM_AP_Password + i, KeyboardBuffer[i]);
      Serial.print(char(EEPROM.read(EEPROM_AP_Password + i))); 
      i++;
    }
    Serial.println("");
    EEPROM.commit();   // update EEPROM   
    Download_WiFi_Login_Details_From_EEPROM(true);  
  }  // end of   if (Length >= 8)  // if entry is > minimum no of characters
  else
  Serial.println(F("Invalid number of characters (min 8, max 20 required"));
}
 // ---------------------------------------------------------------------------
void Check_Time_Control_Relay1()  
{
   const byte Timer1 = 0;
   const byte Timer2 = 1;
   const boolean Timer_On = 0;
   const boolean Timer_Off = 1;
   const boolean Mins = 0;
   const boolean Hrs = 1;
 // Timer 1
   if(Timer_Enable[0] && Clock_Week_Day != 0)  // skip if relay timers are not enabled or clock not valid (weekday = 0)
   {
     if (Relay1_Time_Control_Day_Flags[Timer1][Clock_Week_Day-1]) 
     {    // The clock day is enabled for timer control so check the clock against the timer settings
       if (digitalRead(Relay_Pin[0]) == Relay_Off)  // Check output status (ON or OFF)
       {         // Output is Off so check for Timer ON control
         if(Clock_Hour == Relay1_Time_Control [Timer1][Timer_On][Hrs]) // check hour
         {       //clock hour matches the timer control hour
           if(Clock_Minute == Relay1_Time_Control [Timer1][Timer_On][Mins]) // check minute
           {     // Day, Hour and minute clock agrees with the timer control setting
             if (Enable_Messages)
             {
               Serial.println(F("Relay 1 switched on by Timer 1"));
               Print_Clock();
             }
             Activate_Relay(0, Switch_On,false);   // Switch relay 1 ON, disable output timer 
             Timer_Active[0] = true;    // set relay 1 timer active flag     
           }   
         }  // end of clock hour matches the timer control hour
      }    // end of  if (digitalRead(Relay_Pin[0] == Relay_Off)
      else  // Relay is ON so check for Timer Off control
      {
        if(Clock_Hour == Relay1_Time_Control [Timer1][Timer_Off][Hrs]) // check hour
        {    // check minute
          if(Clock_Minute == Relay1_Time_Control [Timer1][Timer_Off][Mins]) // check minute
          {  // Day, Hour and minute clock agrees with the timer control setting  
            if (Enable_Messages)
            {
              Serial.println(F("Relay 1 switched off by Timer 1"));
              Print_Clock();
            }
            Activate_Relay(0, Switch_Off,false);   // Switch relay 1 Off, disable output timer
            Timer_Active[0] = false;    // reset relay 1 timer active flag      
          }   // end of  if(Clock_Minute = Relay1_Time_Control [Timer1][Timer_Off][Mins]) // check minute 
        }  // end of  if(Clock_Hour = Relay1_Time_Control [Timer1][Timer_Off][Hrs]) // check hour
      }    // end of else  // Relay is ON so check for Timer Off control  
  }  // end of if (Relay1_Time_Control_Day_Flags[Timer1][Clock_Week_Day-1])

      // Timer 2
  if (Relay1_Time_Control_Day_Flags[Timer2][Clock_Week_Day-1]) 
  {    // The clock day is enabled for timer control so check the clock against the timer settings
    if (digitalRead(Relay_Pin[0]) == Relay_Off)  // Check output status (ON or OFF)
    {         // Output is Off so check for Timer ON control
      if(Clock_Hour == Relay1_Time_Control [Timer2][Timer_On][Hrs]) // check hour
      {       //clock hour matches the timer control hour
         if(Clock_Minute == Relay1_Time_Control [Timer2][Timer_On][Mins]) // check minute
         {     // Day, Hour and minute clock agrees with the timer control setting
           if (Enable_Messages)
           {
             Serial.println(F("Relay 1 switched on by Timer 2"));
             Print_Clock();
           }
           Activate_Relay(0, Switch_On,false);   // Switch relay 1 ON, disable output timer
           Timer_Active[0] = true;    // set relay 1 timer active flag      
         }   
      }  // end of    else // clock hour matches the timer control hour
    }    // end of  if (digitalRead(Relay_Pin[0] == Relay_Off)
    else  // Relay is ON so check for Timer Off control
    {
      if(Clock_Hour == Relay1_Time_Control [Timer2][Timer_Off][Hrs]) // check hour
      {    // check minute
        if(Clock_Minute == Relay1_Time_Control [Timer2][Timer_Off][Mins]) // check minute
        {  // Day, Hour and minute clock agrees with the timer control setting  
          if (Enable_Messages)
          {
            Serial.println(F("Relay 1 switched off by Timer 2"));
            Print_Clock();
          }
          Activate_Relay(0, Switch_Off,false);   // Switch relay 1 Off, disable output timer
          Timer_Active[0] = false;    // reset relay 1 timer active flag      
        }   // end of  if(Clock_Minute = Relay1_Time_Control [Timer2][Timer_Off][Mins]) // check minute 
      }  // end of  if(Clock_Hour = Relay1_Time_Control [Timer2][Timer_Off][Hrs]) // check hour
    }    // end of else  // Relay is ON so check for Timer Off control  
  }  // end of if (Relay1_Time_Control_Day_Flags[Timer2][Clock_Week_Day-1]) 
 }    // end of if(Timer_Enable[0] && Clock_Week_Day != 0)  // skip if relay timers are not enabled or clock not valid (weekday = 0)
}   // end of void Check_Time_Control(byte RelayNo) 
 // ---------------------------------------------------------------------------
void Check_Time_Control_Relay2()  
{
   const byte Timer1 = 0;
   const byte Timer2 = 1;
   const boolean Timer_On = 0;
   const boolean Timer_Off = 1;
   const boolean Mins = 0;
   const boolean Hrs = 1;
 // Timer 1
   if(Timer_Enable[1] && Clock_Week_Day != 0)  // skip if relay timers are not enabled or clock not valid (weekday = 0)
   {
     if (Relay2_Time_Control_Day_Flags[Timer1][Clock_Week_Day-1]) 
     {    // The clock day is enabled for timer control so check the clock against the timer settings
       if (digitalRead(Relay_Pin[1]) == Relay_Off)  // Check output status (ON or OFF)
       {         // Output is Off so check for Timer ON control
         if(Clock_Hour == Relay2_Time_Control [Timer1][Timer_On][Hrs]) // check hour
         {       //clock hour matches the timer control hour
           if(Clock_Minute == Relay2_Time_Control [Timer1][Timer_On][Mins]) // check minute
           {     // Day, Hour and minute clock agrees with the timer control setting
             if (Enable_Messages)
             {
               Serial.println(F("Relay 2 switched on by Timer 1"));
               Print_Clock();
             }
             Activate_Relay(1, Switch_On,false);   // Switch relay 2 ON, disable output timer 
             Timer_Active[1] = true;    // set relay 2 timer active flag    
           }   
         }  // end of    else // clock hour matches the timer control hour
      }    // end of  if (digitalRead(Relay_Pin[1] == Relay_Off)
      else  // Relay is ON so check for Timer Off control
      {
        if(Clock_Hour == Relay2_Time_Control [Timer1][Timer_Off][Hrs]) // check hour
        {    // check minute
          if(Clock_Minute == Relay2_Time_Control [Timer1][Timer_Off][Mins]) // check minute
          {  // Day, Hour and minute clock agrees with the timer control setting  
            if (Enable_Messages)
            {
              Serial.println(F("Relay 2 switched off by Timer 1"));
              Print_Clock();
            }
            Activate_Relay(1, Switch_Off,false);   // Switch relay 2 Off, disable output timer
            Timer_Active [1] = false;    // reset relay 2 timer active flag       
          }   // end of  if(Clock_Minute = Relay2_Time_Control [Timer1][Timer_Off][Mins]) // check minute 
        }  // end of  if(Clock_Hour = Relay2_Time_Control [Timer1][Timer_Off][Hrs]) // check hour
      }    // end of else  // Relay is ON so check for Timer Off control  
  }  // end of if (Relay2_Time_Control_Day_Flags[Timer1][Clock_Week_Day-1])

      // Timer 2
  if (Relay2_Time_Control_Day_Flags[Timer2][Clock_Week_Day-1]) 
  {    // The clock day is enabled for timer control so check the clock against the timer settings
    if (digitalRead(Relay_Pin[1]) == Relay_Off)  // Check output status (ON or OFF)
    {         // Output is Off so check for Timer ON control
      if(Clock_Hour == Relay2_Time_Control [Timer2][Timer_On][Hrs]) // check hour
      {       //clock hour matches the timer control hour
         if(Clock_Minute == Relay2_Time_Control [Timer2][Timer_On][Mins]) // check minute
         {     // Day, Hour and minute clock agrees with the timer control setting
           if (Enable_Messages)
           {
             Serial.println(F("Relay 2 switched on by Timer 2"));
             Print_Clock();
           }
           Activate_Relay(1, Switch_On,false);   // Switch relay 2 ON, disable output timer 
           Timer_Active [1] = true;    // set relay timer active flag      
         }   
      }  // end of    else // clock hour matches the timer control hour
    }    // end of  if (digitalRead(Relay_Pin[1] == Relay_Off)
    else  // Relay is ON so check for Timer Off control
    {
      if(Clock_Hour == Relay2_Time_Control [Timer2][Timer_Off][Hrs]) // check hour
      {    // check minute
        if(Clock_Minute == Relay2_Time_Control [Timer2][Timer_Off][Mins]) // check minute
        {  // Day, Hour and minute clock agrees with the timer control setting  
          if (Enable_Messages)
          {
            Serial.println(F("Relay 2 switched off by Timer 2"));
            Print_Clock();
          }
          Activate_Relay(1, Switch_Off,false);   // Switch relay 2 Off, disable output timer 
          Timer_Active[1] = false;    // set relay timer active flag     
        }   // end of  if(Clock_Minute = Relay2_Time_Control [Timer2][Timer_Off][Mins]) // check minute 
      }  // end of  if(Clock_Hour = Relay2_Time_Control [Timer2][Timer_Off][Hrs]) // check hour
    }    // end of else  // Relay is ON so check for Timer Off control  
  }  // end of if (Relay2_Time_Control_Day_Flags[Timer2][Clock_Week_Day-1]) 
 }    // end of if(Timer_Enable[1] && Clock_Week_Day != 0)  // skip if relay timers are not enabled or clock not valid (weekday = 0)
}   // end of void Check_Time_Control2()
 // ---------------------------------------------------------------------------
void Check_Time_Control_Relay3()  
{
   const byte Timer1 = 0;
   const byte Timer2 = 1;
   const boolean Timer_On = 0;
   const boolean Timer_Off = 1;
   const boolean Mins = 0;
   const boolean Hrs = 1;
 // Timer 1
   if(Timer_Enable[2] && Clock_Week_Day != 0)  // skip if relay timers are not enabled or clock not valid (weekday = 0)
   {
     if (Relay3_Time_Control_Day_Flags[Timer1][Clock_Week_Day-1]) 
     {    // The clock day is enabled for timer control so check the clock against the timer settings
       if (digitalRead(Relay_Pin[2]) == Relay_Off)  // Check output status (ON or OFF)
       {         // Output is Off so check for Timer ON control
         if(Clock_Hour == Relay3_Time_Control [Timer1][Timer_On][Hrs]) // check hour
         {       //clock hour matches the timer control hour
           if(Clock_Minute == Relay3_Time_Control [Timer1][Timer_On][Mins]) // check minute
           {     // Day, Hour and minute clock agrees with the timer control setting
             if (Enable_Messages)
             {
               Serial.println(F("Relay 3 switched on by Timer 1"));
               Print_Clock();
             }
             Activate_Relay(2, Switch_On,false);   // Switch relay 3 ON, disable output timer
             Timer_Active[2] = true;    // set relay timer active flag      
           }   
         }  // end of    else // clock hour matches the timer control hour
      }    // end of  if (digitalRead(Relay_Pin[2] == Relay_Off)
      else  // Relay is ON so check for Timer Off control
      {
        if(Clock_Hour == Relay3_Time_Control [Timer1][Timer_Off][Hrs]) // check hour
        {    // check minute
          if(Clock_Minute == Relay3_Time_Control [Timer1][Timer_Off][Mins]) // check minute
          {  // Day, Hour and minute clock agrees with the timer control setting  
            if (Enable_Messages)
            {
              Serial.println(F("Relay 3 switched off by Timer 1"));
              Print_Clock();
            }
            Activate_Relay(2, Switch_Off,false);   // Switch relay 3 Off, disable output timer
            Timer_Active[2] = false;    // set relay timer active flag      
          }   // end of  if(Clock_Minute = Relay3_Time_Control [Timer1][Timer_Off][Mins]) // check minute 
        }  // end of  if(Clock_Hour = Relay3_Time_Control [Timer1][Timer_Off][Hrs]) // check hour
      }    // end of else  // Relay is ON so check for Timer Off control  
  }  // end of if (Relay3_Time_Control_Day_Flags[Timer1][Clock_Week_Day-1])

      // Timer 2
  if (Relay3_Time_Control_Day_Flags[Timer2][Clock_Week_Day-1]) 
  {    // The clock day is enabled for timer control so check the clock against the timer settings
    if (digitalRead(Relay_Pin[2]) == Relay_Off)  // Check output status (ON or OFF)
    {         // Output is Off so check for Timer ON control
      if(Clock_Hour == Relay3_Time_Control [Timer2][Timer_On][Hrs]) // check hour
      {       //clock hour matches the timer control hour
         if(Clock_Minute == Relay3_Time_Control [Timer2][Timer_On][Mins]) // check minute
         {     // Day, Hour and minute clock agrees with the timer control setting
           if (Enable_Messages)
           {
             Serial.println(F("Relay 3 switched on by Timer 2"));
             Print_Clock();
           }
           Activate_Relay(2, Switch_On,false);   // Switch relay 3 ON, disable output timer 
           Timer_Active [2] = true;    
         }   
      }  // end of    else // clock hour matches the timer control hour
    }    // end of  if (digitalRead(Relay_Pin[2] == Relay_Off)
    else  // Relay is ON so check for Timer Off control
    {
      if(Clock_Hour == Relay3_Time_Control [Timer2][Timer_Off][Hrs]) // check hour
      {    // check minute
        if(Clock_Minute == Relay3_Time_Control [Timer2][Timer_Off][Mins]) // check minute
        {  // Day, Hour and minute clock agrees with the timer control setting  
          if (Enable_Messages)
          {
            Serial.println(F("Relay 3 switched off by Timer 2"));
            Print_Clock();
          }
          Activate_Relay(2, Switch_Off,false);   // Switch relay 3 Off, disable output timer  
          Timer_Active[2] = false;   
        }   // end of  if(Clock_Minute = Relay3_Time_Control [Timer2][Timer_Off][Mins]) // check minute 
      }  // end of  if(Clock_Hour = Relay3_Time_Control [Timer2][Timer_Off][Hrs]) // check hour
    }    // end of else  // Relay is ON so check for Timer Off control  
  }  // end of if (Relay3_Time_Control_Day_Flags[Timer2][Clock_Week_Day-1]) 
 }    // end of if(Timer_Enable[2] && Clock_Week_Day != 0)  // skip if relay timers are not enabled or clock not valid (weekday = 0)
}   // end of void Check_Time_Control3() 
    
    // --------------------------------------------------------------------
void Print_Clock()
{
      Serial.print(F("Soft Time is ")); Serial.print(Clock_Hour); // print the hour
      Serial.print(':');
      if (Clock_Minute < 10 ) Serial.print('0');
      Serial.println(Clock_Minute); // print the minute 
      Serial.print(F("Week day is ")); Serial.println(Clock_Week_Day);     
}
    // --------------------------------------------------------------------   
void Check_Timer_Reset_Clock()   // Check timer to see if an automatic reset is required
{
   if(Timer_Reset_Enable)  // only if enabled 
   {            // Clock is enabled so check for timer start
     if(Clock_Hour == Timer_Reset_Clock_Hrs) // check the clock hour against the Timer_Reset clock start time
     {       //clock hour matches the control hour
       if(Clock_Minute == Timer_Reset_Clock_Mins) // check minute against the Timer_Reset clock start time
       {     // Day, Hour and minute clock agrees with the control setting
         Reset_Device("Timer Reset");
       }  // end of    if(Clock_Minute == Timer_Reset_Clock_Mins) // check minute  
     }  // end of    if(Clock_Hour == Clock_Start_Hrs) // check hour
   }   // end of if(Timer_Reset_Enable)  // only if enabled  
   
}   // end of void Check_Timer_Reset_Clock()
 // ----------------------------------------------------------------------
// const unsigned long Pump_Protection_Start_Up_Delay_Time = 60000; // mS - from switching pump ON, the protection routine is disabled
// const unsigned long Pump_Protection_Dry_Time = 5000;  // mS - time for dry running before pump is switched off
unsigned long Run_Timer_Start;   // Time marker for start of the run timer
unsigned long Run_Timer;   // time that the pump has been running (mS)
unsigned long Dry_Timer;  // time in mS that the pump is running dry
unsigned long Dry_Timer_Start;  // time marker for any time that the pump starts to run dry
const boolean Pump_Wet = LOW;    // Protection input status (AUX input 1) when pump is running OK
const boolean Pump_Dry = !Pump_Wet; 
 // ----------------------------------------------------------------------
void Check_Pump_Health()      //  Check pump status and health.  If running dry for 
{                             //  > than the allowed dry period = alarm and turn the pump off
  if (AUX_Mode[1] == 2)  // If pump protection is enabled
  {
    if (digitalRead(Relay_Pin[0]) == Relay_Off)   // Pump is not running
    {
        Run_Timer_Start = millis();    // update timer stamp in preparation for pump protection         
    }
    else  // pump is running
    {
      Run_Timer = millis() - Run_Timer_Start;   // update run timer
      if (Run_Timer >= Pump_Protection_Start_Up_Delay_Time)   // pump is running longer than startup time 
      {                                         // so start protection
        if(Enable_Messages) Serial.println("Starting dry run monitoring");
        if( digitalRead(AUX_Pin[1]) == Pump_Wet)  // Check AUX input one for wet/dry runniong input
        {                                         // pump is running wet
          Dry_Timer_Start = millis();      // update time stamp for dry run start in preparation for a dry run start  
        }  // end of if( digitalRead(AUX_Pin[1]) == Pump_Wet) 
        else // pump is running dry
        {
          Dry_Timer = millis() - Dry_Timer_Start;   // calc time that the pump is running dry
          if(Dry_Timer >= Pump_Protection_Dry_Time)  // ALARM - pump running dry too long
          {  // Shut pump down and raise alarm
            if (Enable_Messages) Serial.println(F("DRY PUMP ALARM"));
            SwitchOneOff();   // Switch relay 1 off
            Alarm = true;  // set alarm flag
          } // end of if(Dry_timer >= Pump_Protection_Dry_Time)
        }   // end of else // pump is running dry
      }   // end of if (Run_Timer >= Pump_Protection_Start_Up_Delay_Time)
    }   // end of else  // pump is running
  }     // end of if (AUX_Mode[0] == 2)
}       // end of boolean Pump_Health()
// -----------------------------------------------------------
void Kill_Alarm()
{
   Alarm = false;      // Reset alarn enable flag - silences buzzer
   digitalWrite(Buzzer,Buzzer_Off);
}
// ------------------------------------------------------
void Auto_EEPROM_Check()   // test EEPROM Block for correct ID, set EEPROM to default if 1st run
{
  byte Value;
  byte i = 0;
  while (i < 5)
  {
    Value = EEPROM.read(EEPROM_Header_Block + i);
    if (Value != i) 
    {    // if any one of header block not correct, reset all EEPROM & exit
      Serial.println(""); Serial.println(F("EEPROM not initialised"));
      Reset_All_EEPROM();
      break;
    }  // end of if (Value != i) 
    i ++;  
  }  // end of while (i < 5) 
}   // end of void Auto_EEPROM_Check() 

